from __future__ import annotations

import enum
import re
from abc import ABC, abstractmethod
from logging import getLogger
from typing import Any, Iterator

from jinja2 import BaseLoader, Environment
from jinja2.exceptions import UndefinedError

from .config import MappingConfig, MappingConfigMappingTable, StairlightConfig


class TemplateSourceType(enum.Enum):
    """Query template source type"""

    FILE = "File"
    GCS = "GCS"
    REDASH = "Redash"
    DBT = "dbt"
    S3 = "S3"

    def __str__(self):
        return self.name


class Template(ABC):
    """Base query template"""

    def __init__(
        self,
        mapping_config: MappingConfig,
        key: str,
        source_type: TemplateSourceType,
        bucket: str | None = None,
        project: str | None = None,
        default_table_prefix: str | None = None,
        data_source_name: str | None = None,
        query_id: int | None = None,
        project_name: str | None = None,
    ):
        """Query template

        Args:
            mapping_config (dict[str, Any]): Mapping configuration
            key (str): A key of query statement.
            source_type (SourceType): Source type.
            bucket (str, optional):
                A name of object storage bucket where query statements saved.
                Defaults to None.
            project (str, optional):
                A project name of Google Cloud where query statements saved.
                Defaults to None.
            default_table_prefix (str, optional):
                If project or dataset that configured table have are omitted,
                it will be complement this prefix. Defaults to None.
            data_source_name (str, optional):
                Data source name. It's only used in a Redash template.
                Defaults to None.
            query_id (int, optional):
                Query id. It's only used in a Redash template.
                Defaults to None.
            project_name (str, optional):
                Project name. It's only used in a DBT template.
                Defaults to None.
        """
        self._mapping_config = mapping_config
        self.key = key
        self.source_type = source_type
        self.bucket = bucket
        self.project = project
        self.default_table_prefix = default_table_prefix
        self.data_source_name = data_source_name
        self.query_id = query_id
        self.project_name = project_name
        self.uri = ""

    def find_mapped_table_attributes(self) -> Iterator[MappingConfigMappingTable]:
        """Get mapped tables as iterator

        Yields:
            Iterator[dict]: Mapped table attributes
        """
        mapping: Any
        for mapping in self._mapping_config.get_mapping():
            not_found: bool = True
            if mapping.TemplateSourceType in (
                TemplateSourceType.FILE.value,
                TemplateSourceType.DBT.value,
            ):
                if self.key.endswith(mapping.FileSuffix):
                    not_found = False
            elif mapping.TemplateSourceType in (
                TemplateSourceType.GCS.value,
                TemplateSourceType.S3.value,
            ):
                if self.uri == mapping.Uri:
                    not_found = False

            if not_found:
                continue

            for table_attributes in mapping.get_table():
                yield table_attributes
            break

    def is_mapped(self) -> bool:
        """Check if the template is set to mapping configuration

        Returns:
            bool: Is mapped or not
        """
        result = False
        for _ in self.find_mapped_table_attributes():
            result = True
            break
        return result

    @staticmethod
    def detect_jinja_params(template_str: str) -> list:
        """Detect jinja parameters from template string

        Args:
            template_str (str): Template string

        Returns:
            list: Jinja parameters
        """
        jinja_expressions = "".join(
            re.findall("{{[^}]*}}", template_str, re.IGNORECASE)
        )
        return [
            param.strip()
            for param in re.findall("[^{}]+", jinja_expressions, re.IGNORECASE)
        ]

    @staticmethod
    def render_by_base_loader(
        source_type: TemplateSourceType,
        key: str,
        template_str: str,
        params: dict[str, Any],
    ) -> str:
        """Render query string from template string

        Args:
            source_type (str): source type
            key (str): key path
            template_str (str): template string
            params (dict[str, Any]): Jinja parameters

        Raises:
            RenderingTemplateException: class RenderingTemplateException

        Returns:
            str: rendered query string
        """
        try:
            jinja_template = Environment(loader=BaseLoader()).from_string(template_str)
            return jinja_template.render(params)
        except UndefinedError as undefined_error:
            raise RenderingTemplateException(
                (
                    f"{undefined_error.message}, "
                    f"source_type: {source_type}, "
                    f"key: {key}"
                )
            ) from None

    @staticmethod
    def ignore_params_from_template_str(
        template_str: str, ignore_params: list[str]
    ) -> str:
        """ignore parameters from template string

        Args:
            template_str (str): template string
            ignore_params (list[str]): ignore parameters

        Returns:
            str: replaced template string
        """
        if not ignore_params:
            ignore_params = []
        replaced_str = template_str
        for ignore_param in ignore_params:
            replaced_str = replaced_str.replace(
                "{{{{ {} }}}}".format(ignore_param), "ignored"
            )
        return replaced_str

    @abstractmethod
    def get_uri(self) -> str:
        """Get uri"""
        pass

    @abstractmethod
    def get_template_str(self) -> str:
        """Get template strings that read from template source"""
        pass

    def render(self, params: dict[str, Any], ignore_params: list[str] = None) -> str:
        """Render a query statement from a jinja template
        Args:
            params (dict[str, Any]): Jinja parameters
            ignore_params (list[str]): Ignore parameters. Defaults to None.
        Returns:
            str: Query statement
        """
        if not ignore_params:
            ignore_params = []
        template_str = self.get_template_str()
        replaced_template_str = self.ignore_params_from_template_str(
            template_str=template_str,
            ignore_params=ignore_params,
        )
        if params:
            results = self.render_by_base_loader(
                source_type=self.source_type,
                key=self.key,
                template_str=replaced_template_str,
                params=params,
            )
        else:
            results = replaced_template_str
        return results


class RenderingTemplateException(Exception):
    """Exception when failing to render jinja templates.

    Args:
        Exception (_type_): Exception
    """

    def __init__(self, *args: object) -> None:
        super().__init__(*args)

    def __str__(self) -> str:
        return super().__str__()


class TemplateSource(ABC):
    """SQL template source"""

    logger = getLogger(__name__)

    def __init__(
        self,
        stairlight_config: StairlightConfig,
        mapping_config: MappingConfig,
        **kwargs,
    ) -> None:
        """SQL template source

        Args:
            stairlight_config (dict[str, Any]): Stairlight configuration
            mapping_config (dict[str, Any]): Mapping configuration
            source_attributes (dict[str, Any]): Template source attributes
        """
        self._stairlight_config = stairlight_config
        self._mapping_config = mapping_config

    @abstractmethod
    def search_templates(self) -> Iterator[Template]:
        """Search query template files

        Yields:
            Iterator[Template]: Attributes of query template files
        """
        pass

    def is_excluded(self, source_type: TemplateSourceType, key: str) -> bool:
        """Check if the specified file is out of scope

        Args:
            source_type (TemplateSourceType): SQL template source type
            key (str): Query template file key

        Returns:
            bool: Return True if the specified file is out of scope
        """
        result = False
        for exclude in self._stairlight_config.get_exclude():
            if source_type == TemplateSourceType(
                exclude.TemplateSourceType
            ) and re.search(rf"{exclude.Regex}", key):
                result = True
                break
        return result



from __future__ import annotations

import argparse
import json
import textwrap
from typing import Any, Callable

from stairlight.stairlight import ResponseType, StairLight


def command_init(stairlight: StairLight, args: argparse.Namespace) -> str:
    """Execute init command

    Args:
        stairlight (StairLight): Stairlight class
        args (argparse.Namespace): CLI arguments

    Returns:
        str: return messages
    """
    message = ""
    stairlight_template_file = stairlight.init()
    if stairlight_template_file:
        message = stairlight_template_file
    return message


def command_check(stairlight: StairLight, args: argparse.Namespace) -> str:
    """Execute check command

    Args:
        stairlight (StairLight): Stairlight class
        args (argparse.Namespace): CLI arguments

    Returns:
        str: return messages
    """
    message = ""
    mapping_template_file = stairlight.check()
    if mapping_template_file:
        message = mapping_template_file

    return message


def command_up(
    stairlight: StairLight, args: argparse.Namespace
) -> dict[str, Any] | list[dict[str, Any]]:
    """Execute up command

    Args:
        stairlight (StairLight): Stairlight class
        args (argparse.Namespace): CLI arguments

    Returns:
        dict[str, Any] | list[dict[str, Any]]: Upstairs results
    """
    return search(
        func=stairlight.up,
        args=args,
        tables=find_tables_to_search(stairlight=stairlight, args=args),
    )


def command_down(
    stairlight: StairLight, args: argparse.Namespace
) -> dict[str, Any] | list[dict[str, Any]]:
    """Execute down command

    Args:
        stairlight (StairLight): Stairlight class
        args (argparse.Namespace): CLI arguments

    Returns:
        dict[str, Any] | list[dict[str, Any]]: Downstairs results
    """
    return search(
        func=stairlight.down,
        args=args,
        tables=find_tables_to_search(stairlight=stairlight, args=args),
    )


def search(
    func: Callable, args: argparse.Namespace, tables: str | list[str]
) -> dict[str, Any] | list[dict[str, Any]]:
    """Search tables by executing stairlight.up() or stairlight.down()

    Args:
        func (Callable): Either stairlight.up() or stairlight.down()
        args (argparse.Namespace): CLI arguments
        tables (str | list[str]): Tables to search

    Returns:
        dict[str, Any] | list[dict[str, Any]]: Results
    """
    results = []
    for table_name in tables:
        result = func(
            table_name=table_name,
            recursive=args.recursive,
            verbose=args.verbose,
            response_type=args.output,
        )
        if len(tables) > 1:
            results.append(result)
        else:
            return result
    return results


def find_tables_to_search(
    stairlight: StairLight, args: argparse.Namespace
) -> list[str]:
    """Find tables to search

    Args:
        stairlight (StairLight): Stairlight class
        args (argparse.Namespace): CLI arguments

    Returns:
        list[str]: Tables to search
    """
    tables_to_search = []
    if args.table:
        tables_to_search = args.table
    elif args.label:
        tables_to_search = stairlight.find_tables_by_labels(target_labels=args.label)
        if not tables_to_search:
            exit()
    return tables_to_search


def set_general_parser(parser: argparse.ArgumentParser) -> None:
    """Set general arguments

    Args:
        parser (argparse.ArgumentParser): ArgumentParser
    """
    parser.add_argument(
        "-c",
        "--config",
        help="set Stairlight configuration directory",
        type=str,
        default=".",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        help="keep silence",
        action="store_true",
        default=False,
    )


def set_save_load_parser(parser: argparse.ArgumentParser) -> None:
    """Set arguments, '--save' and '--load'

    Args:
        parser (argparse.ArgumentParser): ArgumentParser
    """
    parser.add_argument(
        "--save",
        help=textwrap.dedent(
            """\
            A file path where map results will be saved.
            You can choose from local file system, GCS, S3.
        """
        ),
        type=str,
        default=None,
    )
    parser.add_argument(
        "--load",
        help=textwrap.dedent(
            """\
            A file path where map results are saved.
            You can choose from local file system, GCS, S3.
            It can be specified multiple times.
        """
        ),
        action="append",
    )


def set_search_parser(parser: argparse.ArgumentParser) -> None:
    """Set arguments used by up and down

    Args:
        parser (argparse.ArgumentParser): ArgumentParser
    """
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "-t",
        "--table",
        help=textwrap.dedent(
            """\
            table names that Stairlight searches for, can be specified multiple times.
            e.g. -t PROJECT_a.DATASET_b.TABLE_c -t PROJECT_d.DATASET_e.TABLE_f
        """
        ),
        action="append",
    )
    group.add_argument(
        "-l",
        "--label",
        help=textwrap.dedent(
            """\
            labels set for the table in mapping configuration,
            can be specified multiple times.
            The separator between key and value should be a colon(:).
            e.g. -l key_1:value_1 -l key_2:value_2
        """
        ),
        action="append",
    )
    parser.add_argument(
        "-o",
        "--output",
        help="output type",
        type=str,
        choices=[ResponseType.TABLE.value, ResponseType.FILE.value],
        default=ResponseType.TABLE.value,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="return verbose results",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-r",
        "--recursive",
        help="search recursively",
        action="store_true",
        default=False,
    )


def create_parser() -> argparse.ArgumentParser:
    """Create a argument parser

    Returns:
        argparse.ArgumentParser: ArgumentParser
    """
    description = (
        "An end-to-end data lineage tool, "
        "detects table dependencies by SQL SELECT statements. "
        "Without positional arguments, "
        "return a table dependency map as JSON format."
    )
    parser = argparse.ArgumentParser(description=description)
    set_general_parser(parser=parser)
    set_save_load_parser(parser=parser)

    subparsers = parser.add_subparsers()

    # init
    parser_init = subparsers.add_parser(
        "init", help="create new Stairlight configuration file"
    )
    parser_init.set_defaults(handler=command_init)
    set_general_parser(parser=parser_init)

    # map | check
    parser_check = subparsers.add_parser(
        "map",
        aliases=["check"],
        help="create new configuration file about undefined mappings",
    )
    parser_check.set_defaults(handler=command_check)
    set_general_parser(parser=parser_check)

    # up
    parser_up = subparsers.add_parser(
        "up", help="return upstairs ( table | file ) list"
    )
    parser_up.set_defaults(handler=command_up)
    set_general_parser(parser=parser_up)
    set_save_load_parser(parser=parser_up)
    set_search_parser(parser=parser_up)

    # down
    parser_down = subparsers.add_parser(
        "down", help="return downstairs ( table | file ) list"
    )
    parser_down.set_defaults(handler=command_down)
    set_general_parser(parser=parser_down)
    set_save_load_parser(parser=parser_down)
    set_search_parser(parser=parser_down)

    return parser


def main() -> None:
    """CLI entrypoint"""
    parser = create_parser()
    args = parser.parse_args()
    stairlight = StairLight(
        config_dir=args.config, load_files=args.load, save_file=args.save
    )
    stairlight.create_map()

    result = None
    if hasattr(args, "handler"):
        if args.handler == command_init and stairlight.has_stairlight_config():
            exit(f"'{args.config}/stairlight.y(a)ml' already exists.")
        elif args.handler != command_init and not stairlight.has_stairlight_config():
            exit(f"'{args.config}/stairlight.y(a)ml' is not found.")
        result = args.handler(stairlight, args)
    else:
        if not stairlight.has_stairlight_config():
            exit(f"'{args.config}/stairlight.y(a)ml' is not found.")
        result = stairlight.mapped

    if args.quiet or not result:
        return

    if result and isinstance(result, str):
        print(result)
    else:
        print(json.dumps(result, indent=2))


if __name__ == "__main__":
    main()



from __future__ import annotations

import re
from typing import Any, Iterator

from google.cloud import storage

from ..config import ConfigAttributeNotFoundException, MappingConfig, StairlightConfig
from ..controller import GCS_URI_SCHEME
from ..template import Template, TemplateSource, TemplateSourceType
from .config import StairlightConfigIncludeGcs


class GcsTemplate(Template):
    def __init__(
        self,
        mapping_config: MappingConfig,
        key: str,
        bucket: str | None = None,
        project: str | None = None,
        default_table_prefix: str | None = None,
    ):
        super().__init__(
            mapping_config=mapping_config,
            key=key,
            source_type=TemplateSourceType.GCS,
            bucket=bucket,
            project=project,
            default_table_prefix=default_table_prefix,
        )
        self.uri = self.get_uri()

    def get_uri(self) -> str:
        """Get uri from file path

        Returns:
            str: uri
        """
        return f"{GCS_URI_SCHEME}{self.bucket}/{self.key}"

    def get_template_str(self) -> str:
        """Get template string that read from a file in GCS

        Returns:
            str: Template string
        """
        client = storage.Client(credentials=None, project=self.project)
        bucket = client.get_bucket(self.bucket)
        blob = bucket.blob(self.key)
        return blob.download_as_bytes().decode("utf-8")


class GcsTemplateSource(TemplateSource):
    def __init__(
        self,
        stairlight_config: StairlightConfig,
        mapping_config: MappingConfig,
        include: StairlightConfigIncludeGcs,
    ) -> None:
        super().__init__(
            stairlight_config=stairlight_config,
            mapping_config=mapping_config,
        )
        self._include = include

    def search_templates(self) -> Iterator[Template]:
        """Search SQL template files from GCS

        Yields:
            Iterator[SQLTemplate]: SQL template file attributes
        """
        project = self._include.ProjectId
        bucket_name = self._include.BucketName

        if not bucket_name:
            raise ConfigAttributeNotFoundException(
                f"BucketName is not found. {self._include}"
            )

        client = storage.Client(credentials=None, project=self._include.ProjectId)
        blobs: Any = client.list_blobs(bucket_name)
        for blob in blobs:
            if self.is_skipped(blob=blob):
                self.logger.debug(f"{blob.name} is skipped.")
                continue

            yield GcsTemplate(
                mapping_config=self._mapping_config,
                key=blob.name,
                project=project,
                bucket=bucket_name,
                default_table_prefix=self._include.DefaultTablePrefix,
            )

    def is_skipped(self, blob: Any) -> bool:
        """Check the target path is skipped or not

        Args:
            blob (Any): Blob

        Returns:
            bool: Is skipped or not
        """
        return not re.fullmatch(
            rf"{self._include.Regex}",
            blob.name,
        ) or self.is_excluded(
            source_type=TemplateSourceType(self._include.TemplateSourceType),
            key=blob.name,
        )



from __future__ import annotations

import pathlib
import re
from typing import Iterator

from ..config import ConfigAttributeNotFoundException, MappingConfig, StairlightConfig
from ..template import Template, TemplateSource, TemplateSourceType
from .config import StairlightConfigIncludeFile


class FileTemplate(Template):
    def __init__(
        self,
        mapping_config: MappingConfig,
        key: str,
        default_table_prefix: str | None = None,
    ):
        super().__init__(
            mapping_config=mapping_config,
            key=key,
            source_type=TemplateSourceType.FILE,
            default_table_prefix=default_table_prefix,
        )
        self.uri = self.get_uri()

    def get_uri(self) -> str:
        """Get uri from file path

        Returns:
            str: uri
        """
        return str(pathlib.Path(self.key).resolve())

    def get_template_str(self) -> str:
        """Get template string that read from a file in local file system

        Returns:
            str: Template string
        """
        with open(self.key) as f:
            return f.read()


class FileTemplateSource(TemplateSource):
    def __init__(
        self,
        stairlight_config: StairlightConfig,
        mapping_config: MappingConfig,
        include: StairlightConfigIncludeFile,
    ) -> None:
        super().__init__(
            stairlight_config=stairlight_config,
            mapping_config=mapping_config,
        )
        self._include = include

    def search_templates(self) -> Iterator[Template]:
        """Search SQL template files from local file system

        Yields:
            Iterator[Template]: SQL template file attributes
        """
        if not self._include.FileSystemPath:
            raise ConfigAttributeNotFoundException(
                f"FileSystemPath is not found. {self._include}"
            )

        path_obj = pathlib.Path(self._include.FileSystemPath)
        for p in path_obj.glob("**/*"):
            if self.is_skipped(p=p):
                self.logger.debug(f"{str(p)} is skipped.")
                continue

            yield FileTemplate(
                mapping_config=self._mapping_config,
                key=str(p),
                default_table_prefix=self._include.DefaultTablePrefix,
            )

    def is_skipped(self, p: pathlib.Path):
        """Check the target path is skipped or not

        Args:
            p (pathlib.Path): Path

        Returns:
            _type_: Is skipped or not
        """
        return (
            p.is_dir()
            or not re.fullmatch(rf"{self._include.Regex}", str(p))
            or self.is_excluded(
                source_type=TemplateSourceType(self._include.TemplateSourceType),
                key=str(p),
            )
        )



import boto3
from mypy_boto3_s3.service_resource import Object, S3ServiceResource

from ..config_key import S3_URI_SCHEME


def get_s3_object(uri: str) -> Object:
    """Get a S3 object

    Args:
        uri (str): URI

    Returns:
        Object: S3 object
    """
    bucket_name = uri.replace(S3_URI_SCHEME, "").split("/")[0]
    key = uri.replace(f"{S3_URI_SCHEME}{bucket_name}/", "")

    s3: S3ServiceResource = boto3.resource("s3")
    return s3.Object(bucket_name=bucket_name, key=key)



from google.cloud.storage import Blob, Client

from ..config_key import GCS_URI_SCHEME


def get_gcs_blob(uri: str) -> Blob:
    """Get a Google Cloud Storage blob

    Args:
        uri (str): URI

    Returns:
        Blob: Blob
    """
    bucket_name = uri.replace(GCS_URI_SCHEME, "").split("/")[0]
    key = uri.replace(f"{GCS_URI_SCHEME}{bucket_name}/", "")

    client = Client(credentials=None, project=None)
    bucket = client.get_bucket(bucket_name)
    return bucket.blob(key)



from __future__ import annotations

import re
from typing import Iterator

import boto3
from botocore.response import StreamingBody
from mypy_boto3_s3.service_resource import (
    Bucket,
    BucketObjectsCollection,
    ObjectSummary,
    S3ServiceResource,
)
from mypy_boto3_s3.type_defs import GetObjectOutputTypeDef

from ..config import ConfigAttributeNotFoundException, MappingConfig, StairlightConfig
from ..controller import S3_URI_SCHEME
from ..template import Template, TemplateSource, TemplateSourceType
from .config import StairlightConfigIncludeS3


class S3Template(Template):
    def __init__(
        self,
        mapping_config: MappingConfig,
        key: str,
        bucket: str | None = None,
        project: str | None = None,
        default_table_prefix: str | None = None,
    ):
        super().__init__(
            mapping_config=mapping_config,
            key=key,
            source_type=TemplateSourceType.S3,
            bucket=bucket,
            project=project,
            default_table_prefix=default_table_prefix,
        )
        self.uri: str = self.get_uri()
        self.s3: S3ServiceResource = boto3.resource("s3")

    def get_uri(self) -> str:
        """Get uri from file path

        Returns:
            str: uri
        """
        return f"{S3_URI_SCHEME}{self.bucket}/{self.key}"

    def get_template_str(self) -> str:
        """Get template string that read from a file in S3

        Returns:
            str: Template string
        """
        template_str: str = ""
        if not self.bucket:
            return template_str

        s3_bucket: Bucket = self.s3.Bucket(self.bucket)
        object_output: GetObjectOutputTypeDef = s3_bucket.Object(self.key).get()
        body: StreamingBody = object_output["Body"]
        if body:
            template_str = body.read().decode("utf-8")
        return template_str


class S3TemplateSource(TemplateSource):
    def __init__(
        self,
        stairlight_config: StairlightConfig,
        mapping_config: MappingConfig,
        include: StairlightConfigIncludeS3,
    ) -> None:
        super().__init__(
            stairlight_config=stairlight_config,
            mapping_config=mapping_config,
        )
        self._include = include
        self.s3: S3ServiceResource = boto3.resource("s3")

    def search_templates(self) -> Iterator[Template]:
        """Search SQL template files from S3

        Yields:
            Iterator[SQLTemplate]: SQL template file attributes
        """
        project = self._include.ProjectId
        bucket_name = self._include.BucketName

        if not bucket_name:
            raise ConfigAttributeNotFoundException(
                f"BucketName is not found. {self._include}"
            )

        objects: BucketObjectsCollection = self.s3.Bucket(bucket_name).objects.all()
        for obj in objects:
            if self.is_skipped(obj=obj):
                self.logger.debug(f"{obj.key} is skipped.")
                continue

            yield S3Template(
                mapping_config=self._mapping_config,
                key=obj.key,
                project=project,
                bucket=bucket_name,
                default_table_prefix=self._include.DefaultTablePrefix,
            )

    def is_skipped(self, obj: ObjectSummary) -> bool:
        """Check the target object is skipped or not

        Args:
            obj (ObjectSummary): Object

        Returns:
            bool: Is skipped or not
        """
        return not re.fullmatch(
            rf"{self._include.Regex}",
            obj.key,
        ) or self.is_excluded(
            source_type=TemplateSourceType(self._include.TemplateSourceType),
            key=obj.key,
        )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, List, OrderedDict

from ..config import MappingConfigMapping, StairlightConfigInclude
from ..template import TemplateSourceType as source_type


@dataclass
class StairlightConfigIncludeRedash(StairlightConfigInclude):
    TemplateSourceType: str = source_type.REDASH.value
    DatabaseUrlEnvironmentVariable: str = "REDASH_DATABASE_URL"
    DataSourceName: str | None = None
    QueryIds: List[int] = field(default_factory=list)


@dataclass
class MappingConfigMappingRedash(MappingConfigMapping):
    TemplateSourceType: str = source_type.REDASH.value
    QueryId: int | None = None
    DataSourceName: str | None = None
    Tables: List[OrderedDict[str, Any]] = field(default_factory=list)
